;;;; remote-launch-client.scm
;;;; Kon Lovett, Jul '06

;;Issues
;;
;; - Should match version-number w/ server?

(use srfi-1 srfi-9 srfi-18)
(use tcp)
(use openssl miscmacros mathh-int synch)

(eval-when (compile)
  (declare
    (usual-integrations)
    (inline)
    (generic)
    (no-procedure-checks)
    (no-bound-checks)
    (export
    	remote-launch-client-version
    	default-remote-launch-port
    	default-remote-launch-ssl-protocol
			make-remote-launch-client
			remote-launch-client?
			remote-launch-client-name
			remote-launch-client-host
			remote-launch-client-port
			remote-launch-client-context
			remote-platform-request
			remote-repository-list-request
			remote-extension-information-request
			remote-install-request
			remote-uninstall-request
			remote-spawn-request
			remote-exit-request
			remote-shutdown-request) ) )

;;;

(include "version-number")
(include "remote-launch-common")

;;;

(define-parameter default-remote-launch-port DEFAULT-REMOTE-LAUNCH-PORT
	(lambda (x)
		(if (valid-port? x)
			x
			(default-remote-launch-port) ) ) )

(define-parameter default-remote-launch-ssl-protocol 'none
	(lambda (x)
		(if (valid-ssl-protocol-argument? x)
			x
			(default-remote-launch-ssl-protocol) ) ) )

;;;

(define-record-type remote-launch-client
	(%make-remote-launch-client name host port context)
	remote-launch-client?
	(name	remote-launch-client-name)
	(host	remote-launch-client-host)
	(port remote-launch-client-port)
	(context remote-launch-client-context) )

(define-record-printer (remote-launch-client obj port)
	(fprintf port "#<remote-launch-client ~S ~S ~S ~S>"
		(remote-launch-client-name obj)
		(remote-launch-client-host obj)
		(remote-launch-client-port obj)
		(remote-launch-client-context obj)) )

;;;

;;

(define *serial-number* (make-object/synch -1 'serial-number))

(define (next-serial-number)
	(%set!/synch [sno *serial-number*] (add1 sno)) )

;;

(define (make-request-sender client)
	(list #f (remote-launch-client-name client)) )

(define (make-request-packet what client . args)
	`(,what ,(make-request-sender client) ,(next-serial-number) ,@args) )

(define (set-request-packet-sender-address! req addr)
	(set-car! (second req) addr) )

(define (make-platform-request client)
	(make-request-packet 'platform client) )

(define (make-repository-list-request client)
	(make-request-packet 'repository-list client) )

(define (make-extension-information-request client extn)
	(make-request-packet 'extension-information client extn) )

(define (make-install-request client extn host proxy)
	(make-request-packet 'install client extn host proxy) )

(define (make-uninstall-request client extn)
	(make-request-packet 'uninstall client extn) )

(define (make-spawn-request client cmd args)
	(apply make-request-packet 'spawn client cmd args) )

(define (make-exit-request client pid)
	(make-request-packet 'exit client pid) )

(define (make-shutdown-request client pid)
	(make-request-packet 'shutdown client pid) )

;;

(define (response-packet? obj)
	(and (pair? obj)
       (fx<= 4 (length obj))
       (symbol? (response-packet-operation obj))
       (pair? (response-packet-sender obj))
       (string? (response-packet-sender-address obj))
       (fixnum? (response-packet-sender-port obj))
       (pair? (response-packet-receiver obj))
       (string? (response-packet-receiver-address obj))
       (string? (response-packet-receiver-name obj))
       (integer? (response-packet-serial obj)) ) )

;;

(define (request-response client req)
	(let ([ctx (remote-launch-client-context client)]
				[host (remote-launch-client-host client)]
				[port (remote-launch-client-port client)])
		;
		(let*-values (([in out]
										(if ctx
											(ssl-connect host port ctx)
											(tcp-connect host port)))
									([local remote] (tcp-addresses out)))
		;
		(set-request-packet-sender-address! req local)
		(write req out)
		;
		(let ([rsp (read in)])
			;
			(unless (response-packet? rsp)
				(error 'remote-launch-response
					"invalid response packet" rsp))
			(unless (= (request-packet-serial req) (response-packet-serial rsp))
				(error 'remote-launch-response
					"request/response serial number mismatch"
					(request-packet-serial req) (response-packet-serial rsp)))
			;
			(let ([args (response-packet-arguments rsp)])
				(if (error-response-packet? rsp)
					(apply error 'remote-launch-response (car args) (cdr args))
					args ) ) ) ) ) )

;;;

(define (check-pid obj loc)
	(unless (and (fixnum? obj) (fx< 0 obj))
		(error loc "invalid pid" obj)) )

(define (check-extension-name obj loc)
	(unless (or (string? obj) (symbol? obj))
		(error loc "invalid extension name" obj)) )

;;;

;;

(define (remote-launch-client-version)
	version-number )

;;

(define (make-remote-launch-client name #!optional (host "localhost") (port (default-remote-launch-port)) (ctx (default-remote-launch-ssl-protocol)))
	;
	(when (symbol? ctx)
		(set! ctx (and (not (eq? 'none ctx)) (ssl-make-client-context ctx))))
	;
	(unless (or (string? name) (symbol? name))
		(error 'make-remote-launch-client "invalid client name" name))
	(unless (or (not ctx) (ssl-client-context? ctx))
		(error 'make-remote-launch-client "invalid SSL context" ctx))
	;
	(when ctx
		(ssl-load-certificate-chain! ctx *cert-chain-pathname*)
		(ssl-load-certificate-chain! ctx *certs-pathname*)
		(ssl-load-private-key! ctx *priv-key-pathname* #t)
		(ssl-load-verify-root-certificates! ctx #f *cert-directory*)
		(ssl-set-verify! ctx #t))
	;
	(%make-remote-launch-client (->string name) host port ctx) )

;;

(define (remote-platform-request client)
	(request-response client (make-platform-request client)) )

;;

(define (remote-repository-list-request client)
	(request-response client (make-repository-list-request client)) )

;;

(define (remote-extension-information-request client extn)
	(check-extension-name extn 'remote-extension-information-request)
	(request-response client (make-extension-information-request client (->string extn))) )

;;

(define (remote-install-request client extn #!optional (host #f) (proxy #f))
	;
	(check-extension-name extn 'remote-install-request)
	(unless (or (not host) (string? host))
		(error 'remote-install-request "invalid host specification" host))
	(unless (or (not proxy) (string? proxy))
		(error 'remote-install-request "invalid proxy specification" proxy))
	;
	(request-response client (make-install-request client (->string extn) host proxy)) )

;;

(define (remote-uninstall-request client extn)
	(check-extension-name extn 'remote-uninstall-request)
	(request-response client (make-uninstall-request client (->string extn))) )

;;

(define (remote-spawn-request client cmd #!optional (args '()))
	;
	(unless (string? cmd)
		(error 'remote-spawn-request "invalid command name" cmd))
	;
	(request-response client (make-spawn-request client cmd (map ->string args))) )

;;

(define (remote-exit-request client pid)
	(check-pid pid 'remote-exit-request)
	(request-response client (make-exit-request client pid)) )

;;

(define (remote-shutdown-request client pid)
	(check-pid pid 'remote-shutdown-request)
	(request-response client (make-shutdown-request client pid)) )
