;;;; remote-launch-server.scm
;;;; Kon Lovett, Jul '06

;; Issues
;;
;; - Should have server log.
;;
;; - Need to implement a client-identity (via certificate) -> operation-permission
;; model
;;
;;    - Use SSL_get_peer_certificate, extract info to separate struct, then X509_free?
;;    - What info is needed for identity? The hash?
;;
;; - No support for Chicken bundles, all extensions must be via the repo.

;; Overview
;;
;; - Performs local machine (host) Chicken operations as requested
;; by a remote machine (client)
;;
;; - Provides a Capability Security Model based on client identity
;; via OpenSSL Certificate
;;
;; - Provides encrypted packet transport via OpenSLL
;;
;; - Queries:
;;
;;    - Chicken platform: server version, chicken version, processor, byteorder, OS type, and OS version
;;
;;    - installed Chicken extensions
;;
;;    - installed Chicken extension information
;;
;; - Actions:
;;
;;    - start (detached) Chicken process
;;
;;    - stop Chicken process
;;
;;    - install Chicken extension
;;
;;    - uninstall Chicken extension
;;
;;    - install Chicken bundle (*.egg file)
;;
;; - Uses 'chicken-setup' for needed extensions
;;
;; - A Chicken process command (a Script or Program) must be represented as a
;; Chicken extension in the Chicken repository

(use srfi-1 srfi-13 srfi-18)
(use posix tcp)
(use srfi-37 args)
(use openssl tcp-server)

(declare
  (usual-integrations)
  (inline)
  (generic)
  (always-bound
    options) )

;;;

(include "version-number")
(include "remote-launch-common")
(include "launch-setup")
(include "launch-process")

;;;

(define listen-port DEFAULT-REMOTE-LAUNCH-PORT)
(define listen-backlog 4)
(define listen-host #f)
(define ssl-protocol #f)
(define debug-flag #f)

(define (valid-host? obj)
  #t )

(define (valid-backlog? obj)
  (and (fixnum? obj) (fx< 0 obj)) )

;;; Commandline Options

(define (checked-string->number str)
  (handle-exceptions exp #f (string->number str)) )

(define (checked-string->symbol str)
  (handle-exceptions exp #f (string->symbol str)) )

(define (usage)
  (with-output-to-port (current-error-port)
    (lambda ()
      (print "Usage: " (car (argv)) " [options...]")
      (newline)
      (print (args:usage options))
      (print "Report bugs to klovett at pacbell.net")))
  (exit 1) )

(define (usage-error msg)
  (display "Error: " (current-error-port))
  (display msg (current-error-port))
  (newline (current-error-port))
  (usage) )

(define (option-usage-error name msg arg)
  (usage-error (sprintf "option '~A' - ~A: ~A" name msg arg)) )

(define (check-host-arg name arg)
  (unless (valid-host? arg)
    (option-usage-error name "invalid host" arg)) )

(define (check-port-arg name arg)
  (unless (valid-port? (checked-string->number arg))
    (option-usage-error name "invalid port number" arg)) )

(define (check-backlog-arg name arg)
  (unless (valid-backlog? (checked-string->number arg))
    (option-usage-error name "invalid backup amount" arg)) )

(define (check-ssl-protocol-arg name arg)
  (unless (valid-ssl-protocol-argument? (checked-string->symbol arg))
    (option-usage-error name "invalid ssl protocol" arg)) )

(define options `(

  ,(args:make-option (b backlog) (#:required "BACKLOG")
    (conc "number of maximally pending connections [default: " listen-backlog "]")
    (check-backlog-arg name arg)
    (set! listen-backlog (string->symbol arg)) )

  ,(args:make-option (s ssl) (#:required "SSL-PROTOCOL")
    (conc "SSL protocol symbol [default: " (or ssl-protocol 'none) "]")
    (check-ssl-protocol-arg name arg)
    (let ([proto (string->symbol arg)])
      (set! ssl-protocol (if (eq? 'none proto) #f proto))) )

  ,(args:make-option (t host) (#:required "HOST")
    "host restriction [default: <none>]"
    (check-host-arg name arg)
    (set! listen-host (string->symbol arg)) )

  ,(args:make-option (p port) (#:required "PORT")
    (conc "port number [default: " listen-port "]" )
    (check-port-arg name arg)
    (set! listen-port (string->number arg)) )

  ,(args:make-option (d debug) #:none
    "Debug mode"
    (set! debug-flag #t) )

  ,(args:make-option (V version) #:none
    "Display version"
    (print version-number)
    (exit) )

  ,(args:make-option (h help) #:none
    "Display this text"
    (usage) ) ) )

#;
(define-macro (option-set! NAME VAR EXPR)
  `(and-let* ([arg (alist-ref ',NAME opts)])
    (set! ,VAR ,EXPR) ) )

#;
(define-macro (unless-option-set! NAME VAR EXPR)
  `(and-let* ([arg (alist-ref ',NAME opts)])
    (unless ,VAR
      (set! ,VAR ,EXPR)) ) )

(define (process-options)
  (args:parse (command-line-arguments) options)
  #;(receive (opts oprs) (args:parse (command-line-arguments) options)
    (void) ) )

;;;

(define-constant LAUNCH-SERVER-BUFFER-SIZE 4096)

(define launch-listener #f)
(define launch-listener-port #f)

(define (setup-tcp-server)
  (if ssl-protocol
    (begin
      (tcp-server-accept-connection-procedure ssl-accept)
      (tcp-server-prepare-hard-close-procedure ssl-close)
      (set! launch-listener (ssl-listen listen-port listen-backlog listen-host ssl-protocol))
      (set! launch-listener-port (ssl-listener-port launch-listener))
      (ssl-load-certificate-chain! launch-listener *cert-chain-pathname*)
      (ssl-load-private-key! launch-listener *priv-key-pathname* #t)
      (ssl-load-verify-root-certificates! launch-listener #f *cert-directory*)
      (ssl-load-suggested-certificate-authorities! launch-listener *certs-pathname*)
      (ssl-set-verify! launch-listener #t))
    (begin
      (set! launch-listener (tcp-listen listen-port listen-backlog listen-host))
      (set! launch-listener-port (tcp-listener-port launch-listener))))
  (tcp-buffer-size LAUNCH-SERVER-BUFFER-SIZE) )

;;;

(define (platform-information)
  `((server-version       . ,version-number)
    (chicken-version      . ,(chicken-version))
    (machine-byte-order   . ,(machine-byte-order))
    (machine-type         . ,(machine-type))
    (software-type        . ,(software-type))
    (software-version     . ,(software-version)) ) )

;;;

;;

(define (request-packet? obj)
  (and (pair? obj)
       (fx<= 3 (length obj))
       (symbol? (request-packet-operation obj))
       (pair? (request-packet-sender obj))
       (string? (request-packet-sender-address obj))
       (string? (request-packet-sender-name obj))
       (integer? (request-packet-serial obj)) ) )

;;

(define (request-kind? req knd)
  (eq? (request-packet-operation req) knd) )

;;

(define (platform-request? req)
  (request-kind? req 'platform) )

(define (repository-list-request? req)
  (request-kind? req 'repository-list) )

(define (extension-information-request? req)
  (request-kind? req 'extension-information) )

(define (install-request? req)
  (request-kind? req 'install) )

(define (uninstall-request? req)
  (request-kind? req 'uninstall) )

(define (spawn-request? req)
  (request-kind? req 'spawn) )

(define (exit-request? req)
  (request-kind? req 'exit) )

(define (shutdown-request? req)
  (request-kind? req 'shutdown) )

;;

(define (launch-server-identity addr)
  (list addr launch-listener-port) )

(define (make-response-packet req addr what . args)
  `(,what
    ,(launch-server-identity addr)
    ,(if (request-packet? req) (request-packet-sender req) '())
    ,(if (request-packet? req) (request-packet-serial req) '())
    ,@args) )

(define (make-request-response-packet req addr . args)
  (apply make-response-packet req addr (request-packet-operation req) args) )

(define (make-error-response-packet req addr msg . args)
  (apply make-response-packet req addr 'error msg args) )

(define (request-install-host-proxy-values args)
  (values
    (or (and (not-null? args) (first args))
        (chicken-setup-host))
    (or (and (not-null? args) (not-null? (cdr args)) (second args))
        (chicken-setup-proxy))) )

;;

(define (do-platform-request req local)
  (apply make-request-response-packet req local (platform-information)) )

(define (do-repository-list-request req local)
  (let ([out+err (chicken-repository-list)])
    (make-request-response-packet req local (car out+err) (cdr out+err)) ) )

(define (do-extension-information-request req local)
  (let* ([args (request-packet-arguments req)]
         [extn (first args)]
         [info (extension-information extn)])
    (apply make-request-response-packet req local info) ) )

(define (do-install-request req local)
  (let* ([args (request-packet-arguments req)]
         [extn (first args)]
         [out+err
          (receive [host proxy] (request-install-host-proxy-values (cdr args))
            (parameterize ([chicken-setup-host host] [chicken-setup-proxy proxy])
              (chicken-install-extension extn) ) )])
    (make-request-response-packet req local (car out+err) (cdr out+err)) ) )

(define (do-uninstall-request req local)
  (let* ([args (request-packet-arguments req)]
         [extn (first args)]
         [out+err (chicken-uninstall-extension extn)])
    (make-request-response-packet req local (car out+err) (cdr out+err)) ) )

(define (do-spawn-request req local)
  (let* ([args (request-packet-arguments req)]
         [pid (start-process! (first args) (cdr args))])
    (if pid
      (make-request-response-packet req local pid (process-started? pid))
      (make-error-response-packet req local
                                  "cannot spawn process"
                                  (last-process-exception)) ) ) )

(define (do-exit-request req local)
  (let* ([args (request-packet-arguments req)]
         [pid (first args)]
         [term (reap-process! pid)]
         [exp (last-process-exception)])
    (if (or term (not exp))
      (make-request-response-packet req local pid term)
      (make-error-response-packet req local
                                  "cannot check exit status for process"
                                  exp) ) ) )

(define (do-shutdown-request req local)
  (let* ([args (request-packet-arguments req)]
         [pid (first args)]
         [ok (shutdown-process! pid)])
    (if ok
      (make-request-response-packet req local pid (reap-process! pid))
      (make-error-response-packet req local
                                  "cannot terminate process"
                                  (last-process-exception)) ) ) )

;;;

(define (do-request req local remote)
  (cond
    [(platform-request? req)                (do-platform-request req local)]
    [(repository-list-request? req)         (do-repository-list-request req local)]
    [(extension-information-request? req)   (do-extension-information-request req local)]
    [(install-request? req)                 (do-install-request req local)]
    [(uninstall-request? req)               (do-uninstall-request req local)]
    [(spawn-request? req)                   (do-spawn-request req local)]
    [(exit-request? req)                    (do-exit-request req local)]
    [(shutdown-request? req)                (do-shutdown-request req local)]
    [else
      (make-error-response-packet req local
                                  "unrecognized launch request"
                                  (request-packet-operation req))]) )

(define (process-request req local remote)
  (cond
    [(not (request-packet? req))
      (make-error-response-packet req local
                                  "invalid launch request packet"
                                  req)]
    [(not (equal? remote (request-packet-sender-address req)))
      (make-error-response-packet req local
                                  "launch request sender mismatch"
                                  remote (request-packet-sender-address req))]
    [else
      (do-request req local remote)] ) )

(define (launch-server)
  (let ([req (read)])
    (receive [local remote] (tcp-addresses (current-input-port))
      (write (process-request req local remote)) )
    (flush-output) ) )

;;; Main

(process-options)
(setup-tcp-server)
((make-tcp-server launch-listener launch-server) debug-flag)
